/*
 * Copyright (c) 2020 Peter Johanson
 *
 * SPDX-License-Identifier: MIT
 */

#define DT_DRV_COMPAT alps_ec11

#include <device.h>
#include <drivers/gpio.h>
#include <sys/util.h>
#include <kernel.h>
#include <drivers/sensor.h>
#include <sys/__assert.h>
#include <logging/log.h>

#include "ec11.h"

LOG_MODULE_REGISTER(EC11, CONFIG_SENSOR_LOG_LEVEL);

static int ec11_get_ab_state(struct device *dev)
{
	struct ec11_data *drv_data = dev->driver_data;
	const struct ec11_config *drv_cfg = dev->config_info;

	return (gpio_pin_get(drv_data->a, drv_cfg->a_pin) << 1) | gpio_pin_get(drv_data->b, drv_cfg->b_pin);
}

static int ec11_sample_fetch(struct device *dev, enum sensor_channel chan)
{
	struct ec11_data *drv_data = dev->driver_data;
	const struct ec11_config *drv_cfg = dev->config_info;
	u8_t val;
	s8_t delta;

	__ASSERT_NO_MSG(chan == SENSOR_CHAN_ALL || chan == SENSOR_CHAN_ROTATION);

	val = ec11_get_ab_state(dev);

	LOG_DBG("prev: %d, new: %d", drv_data->ab_state, val);

	switch(val | (drv_data->ab_state << 2)) {
		case 0b0010: case 0b0100: case 0b1101: case 0b1011:
			delta = 1;
			break;
		case 0b0001: case 0b0111: case 0b1110: case 0b1000:
			delta = -1;
			break;
		default:
			delta = 0;
			break;
	}

	LOG_DBG("Delta: %d", delta);

	drv_data->pulses += delta;
	drv_data->ab_state = val;

	drv_data->ticks = drv_data->pulses / drv_cfg->resolution;
	drv_data->delta = delta;
	drv_data->pulses %= drv_cfg->resolution;
	
	return 0;
}

static int ec11_channel_get(struct device *dev,
			       enum sensor_channel chan,
			       struct sensor_value *val)
{
	struct ec11_data *drv_data = dev->driver_data;
	
	if (chan != SENSOR_CHAN_ROTATION) {
		return -ENOTSUP;
	}

	val->val1 = drv_data->ticks;
	val->val2 = drv_data->delta;
	
	return 0;
}

static const struct sensor_driver_api ec11_driver_api = {
#ifdef CONFIG_EC11_TRIGGER
	.trigger_set = ec11_trigger_set,
#endif
	.sample_fetch = ec11_sample_fetch,
	.channel_get = ec11_channel_get,
};

int ec11_init(struct device *dev)
{
	struct ec11_data *drv_data = dev->driver_data;
	const struct ec11_config *drv_cfg = dev->config_info;

	LOG_DBG("resolution %d", drv_cfg->resolution);

	drv_data->a = device_get_binding(drv_cfg->a_label);
	if (drv_data->a == NULL) {
		LOG_ERR("Failed to get pointer to A GPIO device");
		return -EINVAL;
	}

	drv_data->b = device_get_binding(drv_cfg->b_label);
	if (drv_data->b == NULL) {
		LOG_ERR("Failed to get pointer to B GPIO device");
		return -EINVAL;
	}

#ifdef CONFIG_EC11_TRIGGER
	if (ec11_init_interrupt(dev) < 0) {
		LOG_DBG("Failed to initialize interrupt!");
		return -EIO;
	}
#endif

	drv_data->ab_state = ec11_get_ab_state(dev);

	return 0;
}

struct ec11_data ec11_data;

const struct ec11_config ec11_cfg = {
	.a_label = DT_INST_GPIO_LABEL(0, a_gpios),
	.a_pin = DT_INST_GPIO_PIN(0, a_gpios),
	.a_flags = DT_INST_GPIO_FLAGS(0, a_gpios),
	.b_label = DT_INST_GPIO_LABEL(0, b_gpios),
	.b_pin = DT_INST_GPIO_PIN(0, b_gpios),
	.b_flags = DT_INST_GPIO_FLAGS(0, b_gpios),
	COND_CODE_0(DT_INST_NODE_HAS_PROP(0, resolution), (1), (DT_INST_PROP(0, resolution))),
};

DEVICE_AND_API_INIT(ec11, DT_INST_LABEL(0), ec11_init,
		    &ec11_data,
		    &ec11_cfg, POST_KERNEL, CONFIG_SENSOR_INIT_PRIORITY,
		    &ec11_driver_api);
